"""
Template module.
"""

import os
import logging
import boto3
import click
import crayons
from botocore.exceptions import ClientError

S3_CLIENT = boto3.client('s3')
CFN_CLIENT = boto3.client('cloudformation')
CFN_TEMPLATE_SIZE_LIMIT = 51200

logging.getLogger('botocore').setLevel(logging.WARNING)


class Template(object):
    """
    CloudFormation template.
    """

    key = None

    def __init__(self, file_path, config):
        self.local_file_path = file_path
        self.file_path = file_path
        self.s3_bucket = config['s3_bucket']
        local_path = config.get('local_path', '')
        if local_path != '.':
            self.file_path = self.file_path.replace(local_path, '')
        self.s3_key = os.path.normpath('{0}/{1}'.format(
            config.get('s3_path', 'cloudformation'),
            self.file_path
        )).strip('/')
        self.public_url = self._public_url()

    def _public_url(self):
        s3_url = os.path.normpath('{0}.s3.amazonaws.com/{1}'.format(self.s3_bucket, self.s3_key))
        return 'https://{0}'.format(s3_url)

    def _content(self):
        try:
            with open(self.local_file_path, 'r') as _file:
                return _file.read()
        except IOError as err:
            click.echo(crayons.red('File {!r} does not exist').format(self.local_file_path), err=True)
            raise err

    def validate(self):
        """
        Validate the template on CloudFormation.

        If the template is larger than CFN_TEMPLATE_SIZE_LIMIT, brume uploads a
        copy of the template with the .copy suffix and performs validation on
        this template.
        """
        local_file_path = self.local_file_path
        validation_path = local_file_path
        params = {'TemplateBody': self._content()}
        if os.path.getsize(self.local_file_path) > CFN_TEMPLATE_SIZE_LIMIT:
            self.upload(copy=True)
            local_file_path = self.local_file_path + '.copy'
            validation_path = self.public_url + '.copy'
            params = {'TemplateURL': validation_path}
        try:
            click.echo('Validating {0} ...'.format(crayons.yellow(validation_path)), nl=False)
            CFN_CLIENT.validate_template(**params)
        except ClientError as err:
            click.echo(crayons.red('invalid'), err=True)
            click.echo(err.message, err=True)
            exit(1)
        else:
            click.echo(crayons.green('valid'))
        return self

    def upload(self, copy=False):
        """
        Upload the template to S3.
        If copy is True,
        """
        s3_key = self.s3_key
        public_url = self.public_url
        if copy:
            s3_key += '.copy'
            public_url += '.copy'
        click.echo('Publishing {0} to {1}'.format(crayons.yellow(self.local_file_path), public_url))
        S3_CLIENT.put_object(Bucket=self.s3_bucket, Body=self._content(), Key=s3_key)
        return self



import unittest
import boto3
from moto import mock_s3

from brume.template import Template

CONFIG = {
    'local_path': 'test_stack',
    's3_bucket': 'dummy-bucket',
    's3_path': 'cloudformation'
}

CFN_TEMPLATE = """{
  "AWSTemplateFormatVersion": "2010-09-09",
  "Description": "test-dev-1: S3",
  "Resources": {
    "MyBucket": {
      "Properties": {
        "BucketName": "my-bucket-test-dev-1",
      },
      "Type": "AWS::S3::Bucket"
    }
  }
}
"""

class TestTemplate(unittest.TestCase):
    """
    Test for brume.Template
    """

    @mock_s3
    def test_upload(self):
        """
        A template can be uploaded to S3.
        """
        conn = boto3.resource('s3')
        conn.create_bucket(Bucket=CONFIG['s3_bucket'])
        template = Template('tests/test_stack/main.json', CONFIG)
        template.upload()
        body = conn.Object(CONFIG['s3_bucket'], 'cloudformation/tests/main.json').get()['Body'].read().decode("utf-8")
        assert body == CFN_TEMPLATE


if __name__ == '__main__':
    unittest.main()



"""
Assets that will be uploaded to S3 (userdata, scripts, configuration files).
"""

import os
import boto3
import click
import crayons

S3_CLIENT = boto3.client('s3')


def send_assets(local_path, s3_bucket, s3_path=''):
    """
    Send directory '{local_path}' under 's3://{s3_bucket}/{s3_path}'
    """
    for (dirpath, _dirnames, filenames) in os.walk(local_path):
        for filename in filenames:
            sourcepath = os.path.join(dirpath, filename)
            key = s3_path + '/' + os.path.relpath(sourcepath, local_path)
            click.echo('Publishing {} to {}'.format(crayons.yellow(sourcepath), s3_bucket + '/' + key))
            with click.open_file(sourcepath) as asset:
                S3_CLIENT.put_object(
                    Bucket=s3_bucket,
                    Body=asset.read(),
                    Key=key)



"""
Brume CLI module.
"""

from glob import glob
from os import path
from yaml import dump

import click
from . import version
from .config import Config
from .stack import Stack
from .template import Template
from .assets import send_assets
from .checker import check_templates


class Context(object):
    """
    Context object to hold the configuration.
    """

    def __init__(self):
        self.config = dict()
        self.stack = None
        self.debug = False


pass_ctx = click.make_pass_decorator(Context, ensure=True)


def config_callback(ctx, _, value):
    ctx = ctx.ensure_object(Context)
    ctx.config = Config.load(value)
    ctx.stack = Stack(ctx.config['stack'])
    return value


@click.group()
@click.version_option(version, '-v', '--version')
@click.help_option('-h', '--help')
@click.option('-c', '--config', expose_value=False, default='brume.yml',
              type=click.File('r'),
              help='Configuration file (defaults to brume.yml).',
              callback=config_callback)
def cli():
    pass


@cli.command()
@pass_ctx
def config(ctx):
    """
    Print the current stack confguration.
    """
    click.echo(dump(ctx.config))


@cli.command()
@pass_ctx
def create(ctx):
    """
    Create a new CloudFormation stack.
    """
    validate_and_upload(ctx.config)
    ctx.stack.create()


@cli.command()
@pass_ctx
def update(ctx):
    """
    Update an existing CloudFormation stack.
    """
    validate_and_upload(ctx.config)
    ctx.stack.update()


@cli.command()
@pass_ctx
def deploy(ctx):
    """
    Create or update a CloudFormation stack.
    """
    validate_and_upload(ctx.config)
    ctx.stack.create_or_update()


@cli.command()
@pass_ctx
def delete(ctx):
    """
    Delete the CloudFormation stack.
    """
    ctx.stack.delete()


@cli.command()
@pass_ctx
def status(ctx):
    """
    Get the status of a CloudFormation stack.
    """
    ctx.stack.status()


@cli.command()
@pass_ctx
def outputs(ctx):
    """
    Get the full list of outputs of a CloudFormation stack.
    """
    click.echo(dump(ctx.stack.outputs(), default_flow_style=False))


@cli.command()
@pass_ctx
def parameters(ctx):
    """
    Get the full list of parameters of a CloudFormation stack.
    """
    for stack_name, stack_parameters in ctx.stack.params().items():
        if not stack_parameters:
            continue
        if ':stack/' in stack_name:
            stack_name = stack_name.partition(':stack/')[2].split('/')[0]
        click.echo(stack_name)
        for param_name, param_value in stack_parameters.items():
            click.echo('\t{} = {}'.format(param_name, param_value))
        click.echo()


@cli.command()
@pass_ctx
def validate(ctx):
    """
    Validate CloudFormation templates.
    """
    return [t.validate() for t in collect_templates(ctx.config)]


@cli.command()
@pass_ctx
def upload(ctx):
    """
    Upload CloudFormation templates and assets to S3.
    """
    process_assets(ctx.config)
    return [t.upload() for t in collect_templates(ctx.config)]


@cli.command()
@pass_ctx
def check(ctx):
    """
    Check CloudFormation templates.
    """
    check_templates(ctx.config['stack']['template_body'])


def process_assets(conf):
    """
    Upload project assets to S3.
    """
    if 'assets' not in conf:
        return
    assets_config = conf['assets']
    local_path = assets_config['local_path']
    s3_bucket = assets_config['s3_bucket']
    s3_path = assets_config['s3_path']
    click.echo("Processing assets from {} to s3://{}/{}".format(local_path, s3_bucket, s3_path))
    send_assets(local_path, s3_bucket, s3_path)


def collect_templates(conf):
    """
    Convert every template into a brume.Template.

    The type of the templates is determined based on the `template_body`
    property of the configuration file.
    """
    _file, ext = path.splitext(conf['stack']['template_body'])
    template_paths = glob(path.join(conf['templates'].get('local_path', ''), '*' + ext))
    return [Template(t, conf['templates']) for t in template_paths]


def validate_and_upload(conf):
    """
    Validate and upload CloudFormation templates to S3.
    """
    templates = collect_templates(conf)
    templates = [t.validate() for t in templates]
    templates = [t.upload() for t in templates]
    process_assets(conf)


if __name__ == '__main__':
    cli()
